local errType = require 'errType'
local utils = require 'utils'
local Network_GetState = require 'methods/Network_GetState'

local Network_SetWifiConfig = {}

local function checkMode(mode)
    if mode == 0 or mode == 2 then
        return true
    end
    return false
end

local function checkKeyMgmt(keyMgmt)
    if keyMgmt == 'NONE' or keyMgmt == 'WPA-PSK' then
        return true
    end
    return false
end

local function enableNetworkService(iface)
    local _, err, _ = utils.shell_posix('sudo', nil, 'systemctl', 'enable', 'wpa_supplicant')
    if err ~= nil then
        return false
    end

    _, err, _ = utils.shell_posix('sudo', nil, 'systemctl', 'enable', 'wpa_supplicant@' .. iface)
    if err ~= nil then
        return false
    end
    return true
end

local function restartNetworkService(iface)
    local _, err, _ = utils.shell_posix('sudo', nil, 'systemctl', 'restart', 'wpa_supplicant')
    if err ~= nil then
        return false
    end

    _, err, _ = utils.shell_posix('sudo', nil, 'systemctl', 'restart', 'wpa_supplicant@' .. iface)
    if err ~= nil then
        return false
    end
    return true
end

function Network_SetWifiConfig.invoke(args)
    local interfaceName, mode, ssid, psk, key_mgmt

    if args ~= nil then
        if utils.isArray(args) then
            interfaceName = args[1]
            mode = args[2]
            ssid = args[3]
            psk = args[4]
            key_mgmt = args[5]
        else
            interfaceName = args.interface
            mode = args.mode
            ssid = args.ssid
            psk = args.psk
            key_mgmt = args.key_mgmt
        end
    end

    if interfaceName == nil or type(interfaceName) ~= 'string' then
        return nil, errType.codes.InvalidParams, {message = 'Failed to read parameter \'interface\'.'}
    end
    if checkMode(mode) == false then
        return nil, errType.codes.InvalidParams, {message = 'Parameter \'mode\' is not valid.'}
    end
    if checkKeyMgmt(key_mgmt) == false then
        return nil, errType.codes.InvalidParams, {message = 'Parameter \'key_mgmt\' is not valid.'}
    end
    if ssid == nil or type(ssid) ~= 'string' then
        return nil, errType.codes.InvalidParams, {message = 'Failed to read parameter \'ssid\'.'}
    end
    if psk ~= nil and type(psk) ~= 'string' then
        return nil, errType.codes.InvalidParams, {message = 'Failed to read parameter \'psk\'.'}
    end

    local interfaces = Network_GetState.invoke()
    if interfaces == nil then
        return nil, errType.codes.InternalError, {message = 'Failed to query interfaces.'}
    end

    local interface
    for _, iface in ipairs(interfaces) do
        if iface.name == interfaceName then
            interface = iface
        end
    end
    if interface == nil then
        return nil, errType.codes.InvalidParams,
        {message = 'Interface \'' .. interfaceName .. '\' not found.'}
    end
    if interface.wireless ~= true then
        return nil, errType.codes.InvalidParams,
        {message = 'Interface \'' .. interfaceName .. '\' is not a wireless interface.'}
    end

    local conf = 'ctrl_interface=DIR=/var/run/wpa_supplicant GROUP=netdev\n'
    conf = conf .. 'network={'
    conf = conf .. '\n\tssid="' .. ssid .. '"'
    conf = conf .. '\n\tmode=' .. mode
    if mode == 0 then
        conf = conf .. '\n\tscan_ssid=1'
    end
    conf = conf .. '\n\tkey_mgmt=' .. key_mgmt
    if key_mgmt == 'WPA-PSK' then
        conf = conf .. '\n\tproto=RSN'
    end
    if psk ~= nil then
        conf = conf .. '\n\tpsk="' .. psk .. '"'
    end
    conf = conf .. '\n}\n'

    local fileName = 'wpa_supplicant-' .. interfaceName .. '.conf'

    local _, err, errData = utils.shell_posix('sudo', conf, 'dd', 'of=/etc/wpa_supplicant/' .. fileName)
    if err ~= nil and errData.exitcode ~= 0 then
        return nil, errType.codes.ServerError,
        {message = 'Failed to write to file \'/etc/wpa_supplicant/' .. fileName .. '\'.'}
    end

    local resEnableSrvc = enableNetworkService(interfaceName)
    local resRestartSrvc = restartNetworkService(interfaceName)

    if resEnableSrvc == false then
        return nil, errType.codes.InternalError, {message = 'Failed to enable network service.'}
    end
    if resRestartSrvc == false then
        return nil, errType.codes.InternalError, {message = 'Failed to restart network service.'}
    end
end

return Network_SetWifiConfig
