local errType = require 'errType'
local utils = require 'utils'

local TimeDate_SetConfig = {}

function TimeDate_SetConfig.isTimeValid(time)
    if time ~= nil then
        if type(time) == 'string' then
            local hours, minutes, seconds = time:match("(%d%d):(%d%d):(%d%d)")
            if hours ~= nil and minutes ~= nil and seconds ~= nil then
                hours = tonumber(hours)
                minutes = tonumber(minutes)
                seconds = tonumber(seconds)

                if hours <= 24 and minutes <= 59 and seconds <= 59 then
                    return true
                end
            end
        end
    end
    return false
end

function TimeDate_SetConfig.isDateValid(date)
    if date ~= nil then
        if type(date) == 'string' then
            local year, month, day = date:match("(%d%d%d%d)%-(%d%d)%-(%d%d)")
            if year ~= nil and month ~= nil and day ~= nil then
                year = tonumber(year)
                month = tonumber(month)
                day = tonumber(day)
                if year > 1900 and year < 2200 and month <= 12 and day <= 31 then
                    return true
                end
            end
        end
    end
    return false
end

function TimeDate_SetConfig.prepareNtpConfFile(file, ntpServers)
    local lines = utils.split(file, '\n')
    local serversLine = table.concat(ntpServers, ' ')

    for i = 1, #lines do
        local currNtpServers = string.match(lines[i], '^Servers%s*=%s*(.*)')
        if currNtpServers ~= nil then
            lines[i] = 'Servers=' .. serversLine
            return table.concat(lines, "\n")
        end
        local currNtp = string.match(lines[i], '^NTP%s*=%s*(.*)')
        if currNtp ~= nil then
            lines[i] = 'NTP=' .. serversLine
            return table.concat(lines, "\n")
        end
    end

    table.insert(lines, 'NTP=' .. serversLine)
    return table.concat(lines, "\n")
end

function TimeDate_SetConfig.invoke(args)
    local time, date, zone, ntp, rtcInLocal, ntpServers

    if args ~= nil then
        if utils.isArray(args) then
            if #args >= 1 then
                date = args[1]
            end
            if #args >= 2 then
                time = args[2]
            end
            if #args >= 3 then
                zone = args[3]
            end
            if #args >= 4 then
                ntp = args[4]
            end
            if #args >= 5 then
                rtcInLocal = args[5]
            end
            if #args >= 6 then
                ntpServers = args[6]
            end
        else
            date = args.date
            time = args.time
            zone = args.zone
            ntp = args.ntp
            rtcInLocal = args.rtcInLocal
            ntpServers = args.ntpServers
        end
    end

    if time ~= nil and TimeDate_SetConfig.isTimeValid(time) == false then
        return nil, errType.codes.InvalidParams,
        {message = 'Failed to read parameter \'time\'.'}
    end
    if date ~= nil and TimeDate_SetConfig.isDateValid(date) == false then
        return nil, errType.codes.InvalidParams,
        {message = 'Failed to read parameter \'date\'.'}
    end
    if ntp ~= nil and (ntp ~= true and ntp ~= false) then
        return nil, errType.codes.InvalidParams,
        {message = 'Failed to read parameter \'ntp\'.'}
    end
    if zone ~= nil and type(zone) ~= 'string' then
        return nil, errType.codes.InvalidParams,
        {message = 'Failed to read parameter \'zone\'.'}
    end
    if rtcInLocal ~= nil and (rtcInLocal ~= true and rtcInLocal ~= false) then
        return nil, errType.codes.InvalidParams,
        {message = 'Failed to read parameter \'rtcInLocal\'.'}
    end
    if ntpServers ~= nil and utils.isArray(ntpServers) == false then
        return nil, errType.codes.InvalidParams,
        {message = 'Failed to read parameter \'ntpServers\'.'}
    end
    if (date ~= nil and time == nil) or (date == nil and time ~= nil) then
        return nil, errType.codes.InvalidParams,
        {message = 'Parameters \'date\' and \'time\' must be both specified.'}
    end

    if rtcInLocal ~= nil then
        local _, err, errData = utils.shell_posix('sudo', nil, 'timedatectl', 'set-local-rtc', tostring(rtcInLocal))
        if err ~= nil then
            return nil, err, errData
        end
    end
    if zone ~= nil then
        local _, err, errData = utils.shell_posix('sudo', nil, 'timedatectl', 'set-timezone', zone)
        if err ~= nil then
            return nil, err, errData
        end
    end
    if ntpServers ~= nil then
        local content = utils.readFromFile('/etc/systemd/timesyncd.conf')
        if content == nil then
            return nil, errType.codes.ServerError,
            {message = 'Failed to read file \'/etc/systemd/timesyncd.conf\'.'}
        end

        local file = TimeDate_SetConfig.prepareNtpConfFile(content, ntpServers)
        local _, err, errData = utils.shell_posix('sudo', file, 'dd', 'of=/etc/systemd/timesyncd.conf')
        if err ~= nil and errData.exitcode ~= 0 then
            return nil, errType.codes.ServerError,
            {message = 'Failed to write to file \'/etc/systemd/timesyncd.conf\'.'}
        end
    end
    if ntp ~= nil then
        local _, err, errData = utils.shell_posix('sudo', nil, 'timedatectl', 'set-ntp', tostring(ntp))
        if err ~= nil then
            return nil, err, errData
        end
    end
    if date ~= nil and time ~= nil then
        local _, err, errData = utils.shell_posix('sudo', nil, 'timedatectl', 'set-time', date .. ' ' .. time)
        if err ~= nil then
            -- check errMsg for timeout, timeout always appears when setting time to lower than current time
            if(errData ~= nil) then
                local timeoutMsg = string.match(errData.message, "error:%s*timeout")
                if timeoutMsg ~= nil then
                    return nil
                end
            end
            return nil, err, errData
        end
    end
    return nil
end

return TimeDate_SetConfig
